from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import collections
import copy
import re

import veriloggen.core.vtypes as vtypes
import veriloggen.core.function as function
import veriloggen.core.task as task
import veriloggen.core.rename_visitor as rename_visitor


class Module(vtypes.VeriloggenNode):
    """ Verilog Module class """

    def __init__(self, name=None, tmp_prefix='_tmp'):
        vtypes.VeriloggenNode.__init__(self)
        self.name = name if name is not None else self.__class__.__name__
        self.io_variable = collections.OrderedDict()
        self.variable = collections.OrderedDict()
        self.global_constant = collections.OrderedDict()
        self.local_constant = collections.OrderedDict()
        self.function = collections.OrderedDict()
        self.task = collections.OrderedDict()
        self.assign = []
        self.always = []
        self.initial = []
        self.instance = collections.OrderedDict()
        self.submodule = collections.OrderedDict()
        self.generate = collections.OrderedDict()
        self.items = []
        self.tmp_prefix = tmp_prefix
        self.tmp_count = 0
        self.hook = []
        self.used = False

    #-------------------------------------------------------------------------
    # User interface for variables
    #-------------------------------------------------------------------------
    def Input(self, name, width=None, length=None, signed=False, value=None):
        t = vtypes.Input(width, length, signed, value, name=name, module=self)
        self.check_existing_identifier(name, vtypes.Wire)
        self.io_variable[name] = t
        self.items.append(t)
        return t

    def Output(self, name, width=None, length=None, signed=False, value=None):
        t = vtypes.Output(width, length, signed, value, name=name, module=self)
        self.check_existing_identifier(name, vtypes.Wire, vtypes.Reg)
        self.io_variable[name] = t
        self.items.append(t)
        return t

    def OutputReg(self, name, width=None, length=None, signed=False, value=None, initval=None):
        t = vtypes.Output(width, length, signed, value, name=name, module=self)
        self.check_existing_identifier(name)
        self.io_variable[name] = t
        self.items.append(t)
        t = vtypes.Reg(width, length, signed, value, initval, name=name)
        self.variable[name] = t
        self.items.append(t)
        return t

    def Inout(self, name, width=None, length=None, signed=False, value=None):
        t = vtypes.Inout(width, length, signed, value, name=name, module=self)
        self.check_existing_identifier(name, vtypes.Wire)
        self.io_variable[name] = t
        self.items.append(t)
        return t

    def Wire(self, name, width=None, length=None, signed=False, value=None):
        t = vtypes.Wire(width, length, signed, value, name=name, module=self)
        self.check_existing_identifier(name, vtypes.Input, vtypes.Output)
        if self.is_reg(name):
            raise ValueError("Object '%s' is already defined." % name)
        self.variable[name] = t
        self.items.append(t)
        return t

    def TmpWire(self, width=None, length=None, signed=False, value=None):
        name = '_'.join([self.tmp_prefix, str(self.get_tmp())])
        return self.Wire(name, width, length, signed, value)

    def Reg(self, name, width=None, length=None, signed=False, value=None, initval=None):
        t = vtypes.Reg(width, length, signed, value,
                       initval, name=name, module=self)
        self.check_existing_identifier(name, vtypes.Output)
        self.variable[name] = t
        self.items.append(t)
        return t

    def TmpReg(self, width=None, length=None, signed=False, value=None, initval=None):
        name = '_'.join([self.tmp_prefix, str(self.get_tmp())])
        return self.Reg(name, width, length, signed, value, initval)

    def Integer(self, name, width=None, length=None, signed=False, value=None, initval=None):
        t = vtypes.Integer(width, length, signed, value,
                           initval, name=name, module=self)
        self.check_existing_identifier(name)
        self.variable[name] = t
        self.items.append(t)
        return t

    def TmpInteger(self, width=None, length=None, signed=False, value=None, initval=None):
        name = '_'.join([self.tmp_prefix, str(self.get_tmp())])
        return self.Integer(name, width, length, signed, value, initval)

    def Real(self, name, width=None, length=None, signed=False, value=None, initval=None):
        t = vtypes.Real(width, length, signed, value,
                        initval, name=name, module=self)
        self.check_existing_identifier(name)
        self.variable[name] = t
        self.items.append(t)
        return t

    def TmpReal(self, width=None, length=None, signed=False, value=None, initval=None):
        name = '_'.join([self.tmp_prefix, str(self.get_tmp())])
        return self.Real(name, width, length, signed, value, initval)

    def Genvar(self, name, width=None, length=None, signed=False, value=None):
        t = vtypes.Genvar(width, length, signed, value, name=name, module=self)
        self.check_existing_identifier(name)
        self.variable[name] = t
        self.items.append(t)
        return t

    def TmpGenvar(self, width=None, length=None, signed=False, value=None):
        name = '_'.join([self.tmp_prefix, str(self.get_tmp())])
        return self.Genvar(name, width, length, signed, value)

    def Parameter(self, name, value, width=None, signed=False, length=None):
        t = vtypes.Parameter(value, width, signed, name=name, module=self)
        self.check_existing_identifier(name)
        self.global_constant[name] = t
        self.items.append(t)
        return t

    def Localparam(self, name, value, width=None, signed=False, length=None):
        t = vtypes.Localparam(value, width, signed, name=name, module=self)
        self.check_existing_identifier(name)
        self.local_constant[name] = t
        self.items.append(t)
        return t

    def TmpLocalparam(self, value, width=None, signed=False, length=None):
        name = '_'.join([self.tmp_prefix, str(self.get_tmp())])
        return self.Localparam(name, value, width, signed, length)

    #-------------------------------------------------------------------------
    def InputLike(self, src, name=None, width=None, length=None, signed=None, value=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        #if length is None: length = src.length
        if length is None:
            length = None
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        return self.Input(name, width, length, signed, value)

    def OutputLike(self, src, name=None, width=None, length=None, signed=None, value=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        #if length is None: length = src.length
        if length is None:
            length = None
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        return self.Output(name, width, length, signed, value)

    def OutputRegLike(self, src, name=None, width=None, length=None, signed=None, value=None, initval=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        #if length is None: length = src.length
        if length is None:
            length = None
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        if initval is None:
            initval = src.initval
        return self.OutputReg(name, width, length, signed, value, initval)

    def InoutLike(self, src, name=None, width=None, length=None, signed=None, value=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        #if length is None: length = src.length
        if length is None:
            length = None
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        return self.Inout(name, width, length, signed, value)

    def WireLike(self, src, name=None, width=None, length=None, signed=None, value=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        return self.Wire(name, width, length, signed, value)

    def TmpWireLike(self, src, width=None, length=None, signed=None, value=None):
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        return self.TmpWire(width, length, signed, value)

    def RegLike(self, src, name=None, width=None, length=None, signed=None, value=None, initval=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        if initval is None:
            initval = src.initval
        return self.Reg(name, width, length, signed, value, initval)

    def TmpRegLike(self, src, width=None, length=None, signed=None, value=None, initval=None):
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        if initval is None:
            initval = src.initval
        return self.TmpReg(width, length, signed, value, initval)

    def IntegerLike(self, src, name=None, width=None, length=None, signed=None, value=None, initval=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        if initval is None:
            initval = src.initval
        return self.Integer(name, width, length, signed, value, initval)

    def TmpIntegerLike(self, src, width=None, length=None, signed=None, value=None, initval=None):
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        if initval is None:
            initval = src.initval
        return self.TmpInteger(width, length, signed, value, initval)

    def RealLike(self, src, name=None, width=None, length=None, signed=None, value=None, initval=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        if initval is None:
            initval = src.initval
        return self.Real(name, width, length, signed, value, initval)

    def TmpRealLike(self, src, width=None, length=None, signed=None, value=None, initval=None):
        if width is None:
            width = src.width
        if length is None:
            length = src.length
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        if initval is None:
            initval = src.initval
        return self.TmpReal(width, length, signed, value, initval)

    def GenvarLike(self, src, name=None, width=None, length=None, signed=None, value=None):
        if name is None:
            name = src.name
        if width is None:
            width = src.width
        #if length is None: length = src.length
        if length is None:
            length = None
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        return self.Genvar(name, width, length, signed, value)

    def TmpGenvarLike(self, src, width=None, length=None, signed=None, value=None):
        if width is None:
            width = src.width
        #if length is None: length = src.length
        if length is None:
            length = None
        if signed is None:
            signed = src.signed
        if value is None:
            value = src.value
        return self.TmpGenvar(width, length, signed, value)

    def ParameterLike(self, src, name=None, value=None, width=None, signed=False, length=None):
        if name is None:
            name = src.name
        if value is None:
            value = src.value
        if width is None:
            width = src.width
        if signed is None:
            signed = src.signed
        if length is None:
            length = src.length
        return self.Parameter(name, value, width, signed, length)

    def LocalparamLike(self, src, name=None, value=None, width=None, signed=False, length=None):
        if name is None:
            name = src.name
        if value is None:
            value = src.value
        if width is None:
            width = src.width
        if signed is None:
            signed = src.signed
        if length is None:
            length = src.length
        return self.Localparam(name, value, width, signed, length)

    def TmpLocalparamLike(self, src, value=None, width=None, signed=False, length=None):
        if value is None:
            value = src.value
        if width is None:
            width = src.width
        if signed is None:
            signed = src.signed
        if length is None:
            length = src.length
        return self.TmpLocalparam(value, width, signed, length)

    #-------------------------------------------------------------------------
    # User interface for control statements
    #-------------------------------------------------------------------------
    def Always(self, *sensitivity):
        t = vtypes.Always(*sensitivity)
        self.always.append(t)
        self.items.append(t)
        return t

    def Assign(self, statement):
        t = vtypes.Assign(statement)
        self.assign.append(t)
        self.items.append(t)
        return t

    def Initial(self, *statement):
        t = vtypes.Initial(*statement)
        self.initial.append(t)
        self.items.append(t)
        return t

    def Function(self, name, width=1):
        t = function.Function(name, width)
        self.check_existing_identifier(name)
        self.function[name] = t
        self.items.append(t)
        return t

    def Task(self, name):
        t = task.Task(name)
        self.check_existing_identifier(name)
        self.task[name] = t
        self.items.append(t)
        return t

    #-------------------------------------------------------------------------
    def GenerateFor(self, pre, cond, post, scope=None):
        t = GenerateFor(self, pre, cond, post, scope)
        if scope is None:
            if None not in self.generate:
                self.generate[None] = []
            self.generate[None].append(t)
            return
        self.check_existing_identifier(scope)
        if scope in self.generate:
            raise ValueError("scope '%s' is already defined." % scope)
        self.generate[scope] = t
        self.items.append(t)
        return t

    def GenerateIf(self, cond, scope=None):
        t = GenerateIf(self, cond, scope)
        if scope is None:
            if None not in self.generate:
                self.generate[None] = []
            self.generate[None].append(t)
            return
        self.check_existing_identifier(scope)
        if scope in self.generate:
            raise ValueError("scope '%s' is already defined." % scope)
        self.generate[scope] = t
        self.items.append(t)
        return t

    #-------------------------------------------------------------------------
    def Instance(self, module, instname, params=None, ports=None):
        if isinstance(module, str):
            module = StubModule(module)
        if not isinstance(module, (Module, StubModule, str)):
            raise TypeError('"module" of Instance must be Module, StubModule, or str, not %s'
                            % type(module))
        self.check_existing_identifier(instname)
        t = Instance(module, instname, params, ports)
        self.instance[instname] = t
        self.items.append(t)
        # if isinstance(module, StubModule):
        #    return None
        if self.find_module(module.name) is not None:
            if self.submodule[module.name] != module:
                raise ValueError(
                    "Module '%s' is already defined." % module.name)
        else:
            self.submodule[module.name] = module
        return t

    #-------------------------------------------------------------------------
    # User intarface for reset assignments
    #-------------------------------------------------------------------------
    def make_reset(self):
        ret = []
        for vname, var in self.variable.items():
            r = var.reset()
            if r is not None:
                ret.append(r)
        return ret

    #-------------------------------------------------------------------------
    # User interface for accessing internal information
    #-------------------------------------------------------------------------
    def get_params(self):
        return self.global_constant

    def get_localparams(self):
        return self.local_constant

    def get_ports(self):
        return self.io_variable

    def get_vars(self):
        return self.variable

    def __getitem__(self, r):
        if isinstance(r, slice):
            raise TypeError("Index must be str.")
        if not isinstance(r, str):
            raise TypeError("Index must be str.")

        v = self.find_identifier(r)
        if v is None:
            raise NameError("No such variable '%s'" % r)
        return v

#    def __getattr__(self, attr):
#        try:
#            return vtypes.VeriloggenNode.__getattr__(self, attr)
#        except AttributeError as e:
#            if attr.startswith('__'):
#                raise e
#
#            if 'object_id' in dir(self):
#                v = self.find_identifier(attr)
#                if v is not None:
#                    return v
#
#            raise e

    #-------------------------------------------------------------------------
    def copy_params(self, src, prefix=None, postfix=None, exclude=None):
        if prefix is None:
            prefix = ''
        if postfix is None:
            postfix = ''
        if exclude is None:
            exclude = ()
        if isinstance(exclude, str):
            exclude = [exclude]
        visitor = rename_visitor.RenameVisitor(prefix, postfix)
        ret = collections.OrderedDict()
        for key, obj in src.global_constant.items():
            skip = False
            for ex in exclude:
                if re.match(ex, key):
                    skip = True
            if skip:
                continue
            copy_obj = copy.deepcopy(obj)
            copy_obj.name = ''.join([prefix, copy_obj.name, postfix])
            copy_obj.width = visitor.visit(copy_obj.width)
            copy_obj.signed = obj.signed
            self.add_object(copy_obj)
            ret[copy_obj.name] = copy_obj
        return ret

    def copy_localparams(self, src, prefix=None, postfix=None, exclude=None):
        if prefix is None:
            prefix = ''
        if postfix is None:
            postfix = ''
        if exclude is None:
            exclude = ()
        if isinstance(exclude, str):
            exclude = [exclude]
        visitor = rename_visitor.RenameVisitor(prefix, postfix)
        ret = collections.OrderedDict()
        for key, obj in src.constant.items():
            skip = False
            for ex in exclude:
                if re.match(ex, key):
                    skip = True
            if skip:
                continue
            copy_obj = copy.deepcopy(obj)
            copy_obj.name = ''.join([prefix, copy_obj.name, postfix])
            copy_obj.width = visitor.visit(copy_obj.width)
            copy_obj.signed = obj.signed
            self.add_object(copy_obj)
            ret[copy_obj.name] = copy_obj
        return ret

    def copy_ports(self, src, prefix=None, postfix=None, exclude=None):
        if prefix is None:
            prefix = ''
        if postfix is None:
            postfix = ''
        if exclude is None:
            exclude = ()
        if isinstance(exclude, str):
            exclude = [exclude]
        visitor = rename_visitor.RenameVisitor(prefix, postfix)
        ret = collections.OrderedDict()
        for key, obj in src.io_variable.items():
            skip = False
            for ex in exclude:
                if re.match(ex, key):
                    skip = True
            if skip:
                continue
            copy_obj = copy.deepcopy(obj)
            copy_obj.name = ''.join([prefix, copy_obj.name, postfix])
            copy_obj.width = visitor.visit(copy_obj.width)
            copy_obj.signed = obj.signed
            self.add_object(copy_obj)
            ret[copy_obj.name] = copy_obj
        return ret

    def copy_vars(self, src, prefix=None, postfix=None, exclude=None):
        if prefix is None:
            prefix = ''
        if postfix is None:
            postfix = ''
        if exclude is None:
            exclude = ()
        if isinstance(exclude, str):
            exclude = [exclude]
        visitor = rename_visitor.RenameVisitor(prefix, postfix)
        ret = collections.OrderedDict()
        for key, obj in src.variable.items():
            skip = False
            for ex in exclude:
                if re.match(ex, key):
                    skip = True
            if skip:
                continue
            copy_obj = copy.deepcopy(obj)
            copy_obj.name = ''.join([prefix, copy_obj.name, postfix])
            copy_obj.width = visitor.visit(copy_obj.width)
            copy_obj.signed = obj.signed
            self.add_object(copy_obj)
            ret[copy_obj.name] = copy_obj
        return ret

    def copy_sim_ports(self, src, prefix=None, postfix=None, exclude=None):
        if prefix is None:
            prefix = ''
        if postfix is None:
            postfix = ''
        if exclude is None:
            exclude = ()
        if isinstance(exclude, str):
            exclude = [exclude]
        visitor = rename_visitor.RenameVisitor(prefix, postfix)
        ret = collections.OrderedDict()
        for key, obj in src.io_variable.items():
            skip = False
            for ex in exclude:
                if re.match(ex, key):
                    skip = True
            if skip:
                continue
            copy_obj = self.get_opposite_variable(obj)(
                name=key, width=copy.deepcopy(obj.width))
            copy_obj.name = ''.join([prefix, copy_obj.name, postfix])
            copy_obj.width = visitor.visit(copy_obj.width)
            copy_obj.signed = obj.signed
            self.add_object(copy_obj)
            ret[copy_obj.name] = copy_obj
        return ret

    #-------------------------------------------------------------------------
    def connect_params(self, targ, prefix=None, postfix=None, exclude=None, strict=False):
        if prefix is None:
            prefix = ''
        if postfix is None:
            postfix = ''
        if exclude is None:
            exclude = ()
        if isinstance(exclude, str):
            exclude = [exclude]
        ret = []
        for key, obj in targ.global_constant.items():
            skip = False
            for ex in exclude:
                if re.match(ex, key):
                    skip = True
            if skip:
                continue
            my_key = ''.join([prefix, key, postfix])
            if strict and (my_key not in self.global_constant) and (my_key not in self.local_constant):
                raise IndexError(
                    "No such constant '%s' in module '%s'" % (key, self.name))
            if my_key in self.global_constant:
                ret.append((key, self.global_constant[my_key]))
            elif my_key in self.local_constant:
                ret.append((key, self.local_constant[my_key]))
        return ret

    def connect_ports(self, targ, prefix=None, postfix=None, exclude=None, strict=False):
        if prefix is None:
            prefix = ''
        if postfix is None:
            postfix = ''
        if exclude is None:
            exclude = ()
        if isinstance(exclude, str):
            exclude = [exclude]
        ret = []
        for key, obj in targ.io_variable.items():
            skip = False
            for ex in exclude:
                if re.match(ex, key):
                    skip = True
            if skip:
                continue
            my_key = ''.join([prefix, key, postfix])
            if strict and (my_key not in self.io_variable) and (my_key not in self.variable):
                raise IndexError("No such IO '%s' in module '%s'" %
                                 (key, self.name))
            if my_key in self.io_variable:
                ret.append((key, self.io_variable[my_key]))
            elif my_key in self.variable:
                ret.append((key, self.variable[my_key]))
        return ret

    #-------------------------------------------------------------------------
    # User interface for Verilog code generation
    #-------------------------------------------------------------------------
    def to_verilog(self, filename=None):
        import veriloggen.verilog.to_verilog as to_verilog
        obj = self.to_hook_resolved_obj()
        return to_verilog.write_verilog(obj, filename)

    def add_hook(self, method, args=None, kwargs=None):
        """ add a hooked method to 'to_verilog()' """
        self.hook.append((method, args, kwargs))

    #-------------------------------------------------------------------------
    def add_object(self, obj):
        if isinstance(obj, vtypes._Variable) and obj.name is None:
            raise ValueError("Object must have a name.")

        self.items.append(obj)

        if isinstance(obj, vtypes.AnyType):
            self.io_variable[obj.name] = obj
            #self.variable[obj.name] = obj
            #self.global_constant[obj.name] = obj
            #self.local_constant[obj.name] = obj
            return

        if isinstance(obj, (vtypes.Input, vtypes.Output, vtypes.Inout)):
            self.io_variable[obj.name] = obj
            return

        if isinstance(obj, (vtypes.Reg, vtypes.Wire)):
            self.variable[obj.name] = obj
            return

        if isinstance(obj, (vtypes.Integer, vtypes.Real, vtypes.Genvar)):
            self.variable[obj.name] = obj
            return

        if isinstance(obj, vtypes.Parameter):
            self.global_constant[obj.name] = obj
            return

        if isinstance(obj, vtypes.Localparam):
            self.local_constant[obj.name] = obj
            return

        if isinstance(obj, function.Function):
            self.function[obj.name] = obj
            return

        if isinstance(obj, task.Task):
            self.task[obj.name] = obj
            return

        if isinstance(obj, vtypes.Assign):
            self.assign.append(obj)
            return

        if isinstance(obj, vtypes.Always):
            self.always.append(obj)
            return

        if isinstance(obj, vtypes.Initial):
            self.initial.append(obj)
            return

        if isinstance(obj, GenerateFor):
            if obj.scope is None:
                if None not in self.generate:
                    self.generate[None] = []
                self.generate[None].append(obj)
                return
            self.generate[obj.scope] = obj
            return

        if isinstance(obj, GenerateIf):
            if obj.true_scope is None:
                if None not in self.generate:
                    self.generate[None] = []
                self.generate[None].append(obj)
                return
            self.generate[obj.true_scope] = obj
            return

        if isinstance(obj, Instance):
            if isinstance(obj.module, Module):
                self.instance[obj.instname] = obj
                self.submodule[obj.module.name] = obj.module
            elif isinstance(obj.module, str):
                self.instance[obj.instname] = obj
            return

        raise TypeError("Object type '%s' is not supported." % str(type(obj)))

    #-------------------------------------------------------------------------
    def add_function(self, t):
        if not isinstance(t, function.Function):
            raise TypeError(
                "add_function requires a Function, not %s" % type(t))
        name = t.name
        self.function[name] = t
        self.items.append(t)
        return t

    def add_task(self, t):
        if not isinstance(t, task.Task):
            raise TypeError("add_task requires a Task, not %s" % type(t))
        name = t.name
        self.task[name] = t
        self.items.append(t)
        return t

    #-------------------------------------------------------------------------
    def find_identifier(self, name):
        if name in self.io_variable:
            return self.io_variable[name]
        if name in self.variable:
            return self.variable[name]
        if name in self.global_constant:
            return self.global_constant[name]
        if name in self.local_constant:
            return self.local_constant[name]
        if name in self.function:
            return self.function[name]
        if name in self.task:
            return self.task[name]
        if name in self.instance:
            return self.instance[name]
        if name in self.generate:
            return self.generate[name]
        return None

    #-------------------------------------------------------------------------
    def get_tmp(self):
        ret = self.tmp_count
        self.tmp_count += 1
        return ret

    #-------------------------------------------------------------------------
    def is_input(self, name):
        if name not in self.io_variable:
            return False
        if isinstance(self.io_variable[name], vtypes.Input):
            return True
        return False

    def is_output(self, name):
        if name not in self.io_variable:
            return False
        if isinstance(self.io_variable[name], vtypes.Output):
            return True
        return False

    def is_inout(self, name):
        if name not in self.io_variable:
            return False
        if isinstance(self.io_variable[name], vtypes.Inout):
            return True
        return False

    def is_reg(self, name):
        if name not in self.variable:
            return False
        if isinstance(self.variable[name], vtypes.Reg):
            return True
        return False

    def is_wire(self, name):
        if name not in self.variable and name not in self.io_variable:
            return False
        if name in self.variable and isinstance(self.variable[name], vtypes.Wire):
            return True
        if name in self.variable and isinstance(self.variable[name], vtypes.Reg):
            return False
        if name in self.io_variable:
            return True
        return False

    #-------------------------------------------------------------------------
    def get_opposite_variable(self, var, use_wire=False):
        if isinstance(var, vtypes.Input):
            if use_wire:
                return vtypes.Wire
            return vtypes.Reg
        if isinstance(var, vtypes.Output):
            return vtypes.Wire
        if isinstance(var, vtypes.Inout):
            return vtypes.Wire
        raise TypeError('No corresponding IO type for %s' % str(type(var)))

    #-------------------------------------------------------------------------
    def to_hook_resolved_obj(self):
        # if there is no hooked method, object copy is not required.
        if not self.has_hook():
            return self
        copied = copy.deepcopy(self)
        copied.resolve_hook()
        return copied

    def resolve_hook(self):
        for method, args, kwargs in self.hook:
            if args is None:
                args = ()
            if kwargs is None:
                kwargs = {}
            method(*args, **kwargs)

        for sub in self.submodule.values():
            sub.resolve_hook()

    def has_hook(self):
        if self.hook:
            return True
        for sub in self.submodule.values():
            if sub.has_hook():
                return True
        return False

    #-------------------------------------------------------------------------
    def find_module(self, name):
        if name in self.submodule:
            return self.submodule[name]
        for gen in self.generate.values():
            if isinstance(gen, (tuple, list)):
                for g in gen:
                    r = g.find_module(name)
                    if r is not None:
                        return r
            else:
                r = gen.find_module(name)
                if r is not None:
                    return r
        for sub in self.submodule.values():
            r = sub.find_module(name)
            if r is not None:
                return r
        return None

    def get_modules(self):
        modules = collections.OrderedDict()
        modules[self.name] = self
        for gen in self.generate.values():
            if isinstance(gen, (tuple, list)):
                for g in gen:
                    modules.update(g.get_modules())
            else:
                modules.update(gen.get_modules())
        for sub in self.submodule.values():
            modules.update(sub.get_modules())
        return modules

    def check_existing_identifier(self, name, *types):
        s = self.find_identifier(name)
        if s is None:
            return
        if isinstance(s, vtypes.AnyType):
            return
        if len(types) == 0:
            raise ValueError("Object '%s' is already defined." % name)
        if not isinstance(s, types):
            raise ValueError("Object '%s' is already defined." % name)


class StubModule(vtypes.VeriloggenNode):
    """ Verilog Module class """

    def __init__(self, name=None, code=''):
        vtypes.VeriloggenNode.__init__(self)
        self.name = name if name is not None else self.__class__.__name__
        self.code = code
        self.used = False

    def set_code(self, code):
        self.code = code

    def get_code(self):
        return self.code

    def to_verilog(self, filename=None):
        import veriloggen.verilog.to_verilog as to_verilog
        return to_verilog.write_verilog(self, filename)

    def resolve_hook(self):
        pass

    def has_hook(self):
        return False

    def find_module(self, name):
        return None

    def get_modules(self):
        modules = collections.OrderedDict()
        modules[self.name] = self
        return modules


class Instance(vtypes.VeriloggenNode):

    def __init__(self, module, instname, params=None, ports=None):
        vtypes.VeriloggenNode.__init__(self)
        if params is None:
            params = ()
        if ports is None:
            ports = ()
        self._type_check_params(params)
        self._type_check_ports(ports)
        self.module = module
        self.instname = instname
        if hasattr(self.module, 'used'):
            self.module.used = True

        if not params:
            self.params = ()
        elif isinstance(params, dict):  # named
            self.params = [(k, v) for k, v in params.items()]
        elif isinstance(params[0], (tuple, list)):  # named
            for param in params:
                if not isinstance(param, (tuple, list)) or len(param) != 2:
                    raise ValueError("Illegal parameter argument")
            self.params = params
        else:  # noname
            for param in params:
                if not isinstance(param, vtypes.numerical_types):
                    raise ValueError("Illegal parameter argument")
            if not isinstance(module, Module) or isinstance(module, StubModule):
                self.params = [(None, p) for p in params]
            else:
                self.params = [(v.name, p) for v, p in zip(
                    module.global_constant.values(), params)]

        if isinstance(module, Module) and not isinstance(module, StubModule):
            for name, port in self.params:
                if name is None:
                    continue
                if not isinstance(module.find_identifier(name), vtypes.Parameter):
                    raise ValueError("No such parameter '%s' in module '%s'" %
                                     (name, module.name))

        if not ports:
            self.ports = ()
        elif isinstance(ports, dict):  # named
            self.ports = [(k, v) for k, v in ports.items()]
        elif isinstance(ports[0], (tuple, list)):  # named
            for port in ports:
                if not isinstance(port, (tuple, list)) or len(port) != 2:
                    raise ValueError("Illegal port argument")
            self.ports = ports
        else:  # noname
            for port in ports:
                if port is not None and not isinstance(port, vtypes.numerical_types):
                    raise ValueError("Illegal port argument")
            if not isinstance(module, Module) or isinstance(module, StubModule):
                self.ports = [(None, p) for p in ports]
            else:
                self.ports = [(v.name, p) for v, p in zip(
                    module.io_variable.values(), ports)]

        if isinstance(module, Module) and not isinstance(module, StubModule):
            for name, port in self.ports:
                if name is None:
                    continue
                if not isinstance(module.find_identifier(name),
                                  (vtypes.Input, vtypes.Output, vtypes.Inout)):
                    raise ValueError("No such port '%s' in module '%s'" %
                                     (name, module.name))

    def _type_check_module(self, module):
        if not isinstance(module, (Module, StubModule)):
            raise TypeError("module of Instance must be Module or StubModule, not %s" %
                            type(module))

    def _type_check_params(self, params):
        if not isinstance(params, (tuple, list, dict)):
            raise TypeError(
                "params of Instance require tuple, list, or dict, not %s." % type(params))

    def _type_check_ports(self, ports):
        if not isinstance(ports, (tuple, list, dict)):
            raise TypeError(
                "ports of Instance require tuple, list, or dict, not %s." % type(ports))


class Generate(Module):
    """ Base class of generate statement """

    def __init__(self, m):
        Module.__init__(self)
        self.m = m

    def Input(self, name, width=None, length=None, signed=False, value=None):
        raise TypeError("Input port is not allowed in generate statement")

    def Output(self, name, width=None, length=None, signed=False, value=None):
        raise TypeError("Output port is not allowed in generate statement")

    def OutputReg(self, name, width=None, length=None, signed=False, value=None, initval=None):
        raise TypeError("OutputReg port is not allowed in generate statement")

    def Inout(self, name, width=None, length=None, signed=False, value=None):
        raise TypeError("Inout port is not allowed in generate statement")

    def find_identifier(self, name):
        r = Module.find_identifier(self, name)
        if r is not None:
            return r
        r = self.m.find_identifier(name)
        if r is not None:
            return r
        return None

    def get_modules(self):
        modules = collections.OrderedDict()
        for gen in self.generate.values():
            if isinstance(gen, (tuple, list)):
                for g in gen:
                    modules.update(g.get_modules())
            else:
                modules.update(gen.get_modules())
        for sub in self.submodule.values():
            modules.update(sub.get_modules())
        return modules

    def _type_check_scope(self, scope):
        if scope is None:
            return
        if not isinstance(scope, str):
            raise TypeError("Scope name should be str, not %s." % type(scope))


class GenerateFor(Generate):

    def __init__(self, m, pre, cond, post, scope=None):
        Generate.__init__(self, m)
        self.pre = pre
        self.cond = cond
        self.post = post
        self.scope = scope
        self._type_check_scope(scope)

    def __getitem__(self, index):
        return vtypes.ScopeIndex(self.scope, index)


class GenerateIf(Generate):

    def __init__(self, m, cond, true_scope=None):
        Generate.__init__(self, m)
        self.cond = cond
        self.true_scope = true_scope
        self.Else = GenerateIfElse(m)
        self._type_check_scope(true_scope)


class GenerateIfElse(Generate):

    def __init__(self, m, false_scope=None):
        Generate.__init__(self, m)
        self.false_scope = false_scope
        self._type_check_scope(false_scope)

    def __call__(self, false_scope):
        self.false_scope = false_scope
        return self


def connect_same_name(*args):
    ret = []
    for arg in args:
        if isinstance(arg, (list, tuple)):
            ret.extend([(a.name, a) for a in arg])
        elif isinstance(arg, vtypes._Variable):
            ret.append((arg.name, arg))
        else:
            raise TypeError(
                'connect_same_name supports Variables, lists and tuples of them.')
    return ret



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import thread_multithread

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [8-1:0] LED;

  blinkled
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #10000;
    $finish;
  end


endmodule



module blinkled
(
  input CLK,
  input RST,
  output reg [8-1:0] LED
);

  reg [8-1:0] count;
  reg [32-1:0] th_countup;
  localparam th_countup_init = 0;
  reg signed [32-1:0] _th_countup_times_0;
  reg signed [32-1:0] _th_countup_inc_1;
  reg signed [32-1:0] _th_countup_i_2;
  reg [32-1:0] th_blink;
  localparam th_blink_init = 0;
  reg signed [32-1:0] _th_blink_times_3;
  reg signed [32-1:0] _th_blink_inc_4;
  reg signed [32-1:0] _th_blink_i_5;
  localparam th_countup_1 = 1;
  localparam th_countup_2 = 2;
  localparam th_countup_3 = 3;
  localparam th_countup_4 = 4;
  localparam th_countup_5 = 5;
  localparam th_countup_6 = 6;
  localparam th_countup_7 = 7;

  always @(posedge CLK) begin
    if(RST) begin
      th_countup <= th_countup_init;
      _th_countup_times_0 <= 0;
      _th_countup_inc_1 <= 0;
      count <= 0;
      _th_countup_i_2 <= 0;
    end else begin
      case(th_countup)
        th_countup_init: begin
          _th_countup_times_0 <= 20;
          _th_countup_inc_1 <= 1;
          th_countup <= th_countup_1;
        end
        th_countup_1: begin
          count <= 0;
          th_countup <= th_countup_2;
        end
        th_countup_2: begin
          _th_countup_i_2 <= 0;
          th_countup <= th_countup_3;
        end
        th_countup_3: begin
          if(_th_countup_i_2 < _th_countup_times_0) begin
            th_countup <= th_countup_4;
          end else begin
            th_countup <= th_countup_7;
          end
        end
        th_countup_4: begin
          count <= count + _th_countup_inc_1;
          th_countup <= th_countup_5;
        end
        th_countup_5: begin
          $display("count = %d", count);
          th_countup <= th_countup_6;
        end
        th_countup_6: begin
          _th_countup_i_2 <= _th_countup_i_2 + 1;
          th_countup <= th_countup_3;
        end
      endcase
    end
  end

  localparam th_blink_1 = 1;
  localparam th_blink_2 = 2;
  localparam th_blink_3 = 3;
  localparam th_blink_4 = 4;
  localparam th_blink_5 = 5;
  localparam th_blink_6 = 6;
  localparam th_blink_7 = 7;

  always @(posedge CLK) begin
    if(RST) begin
      th_blink <= th_blink_init;
      _th_blink_times_3 <= 0;
      _th_blink_inc_4 <= 0;
      LED <= 0;
      _th_blink_i_5 <= 0;
    end else begin
      case(th_blink)
        th_blink_init: begin
          _th_blink_times_3 <= 10;
          _th_blink_inc_4 <= 1;
          th_blink <= th_blink_1;
        end
        th_blink_1: begin
          LED <= 0;
          th_blink <= th_blink_2;
        end
        th_blink_2: begin
          _th_blink_i_5 <= 0;
          th_blink <= th_blink_3;
        end
        th_blink_3: begin
          if(_th_blink_i_5 < _th_blink_times_3) begin
            th_blink <= th_blink_4;
          end else begin
            th_blink <= th_blink_7;
          end
        end
        th_blink_4: begin
          LED <= LED + _th_blink_inc_4;
          th_blink <= th_blink_5;
        end
        th_blink_5: begin
          $display("  led = %d", LED);
          th_blink <= th_blink_6;
        end
        th_blink_6: begin
          _th_blink_i_5 <= _th_blink_i_5 + 1;
          th_blink <= th_blink_3;
        end
      endcase
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = thread_multithread.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import dataflow__iter

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg [4-1:0] xdata;
  reg [4-1:0] ydata;
  wire [4-1:0] zdata;
  reg xvalid;
  wire xready;
  assign xready = 1;
  reg yvalid;
  wire yready;
  assign yready = 1;
  wire zvalid;
  reg zready;
  assign zvalid = 1;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .xdata(xdata),
    .ydata(ydata),
    .zdata(zdata)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    xdata = 0;
    xvalid = 0;
    ydata = 0;
    yvalid = 0;
    zready = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #10000;
    $finish;
  end

  reg [32-1:0] xfsm;
  localparam xfsm_init = 0;
  reg [32-1:0] _tmp_0;
  localparam xfsm_1 = 1;
  localparam xfsm_2 = 2;
  localparam xfsm_3 = 3;
  localparam xfsm_4 = 4;

  always @(posedge CLK) begin
    if(RST) begin
      xfsm <= xfsm_init;
      _tmp_0 <= 0;
    end else begin
      case(xfsm)
        xfsm_init: begin
          xvalid <= 0;
          if(reset_done) begin
            xfsm <= xfsm_1;
          end 
        end
        xfsm_1: begin
          xvalid <= 1;
          xfsm <= xfsm_2;
        end
        xfsm_2: begin
          if(xready) begin
            xdata <= xdata + 1;
          end 
          if(xready) begin
            _tmp_0 <= _tmp_0 + 1;
          end 
          if((_tmp_0 == 5) && xready) begin
            xfsm <= xfsm_3;
          end 
        end
        xfsm_3: begin
          if(xready) begin
            xdata <= xdata + 1;
          end 
          if(xready) begin
            _tmp_0 <= _tmp_0 + 1;
          end 
          if((_tmp_0 == 10) && xready) begin
            xfsm <= xfsm_4;
          end 
        end
      endcase
    end
  end

  reg [32-1:0] yfsm;
  localparam yfsm_init = 0;
  reg [32-1:0] _tmp_1;
  localparam yfsm_1 = 1;
  localparam yfsm_2 = 2;
  localparam yfsm_3 = 3;
  localparam yfsm_4 = 4;

  always @(posedge CLK) begin
    if(RST) begin
      yfsm <= yfsm_init;
      _tmp_1 <= 0;
    end else begin
      case(yfsm)
        yfsm_init: begin
          yvalid <= 0;
          if(reset_done) begin
            yfsm <= yfsm_1;
          end 
        end
        yfsm_1: begin
          yvalid <= 1;
          yfsm <= yfsm_2;
        end
        yfsm_2: begin
          if(yready) begin
            ydata <= ydata + 2;
          end 
          if(yready) begin
            _tmp_1 <= _tmp_1 + 1;
          end 
          if((_tmp_1 == 5) && yready) begin
            yfsm <= yfsm_3;
          end 
        end
        yfsm_3: begin
          if(yready) begin
            ydata <= ydata + 2;
          end 
          if(yready) begin
            _tmp_1 <= _tmp_1 + 1;
          end 
          if((_tmp_1 == 10) && yready) begin
            yfsm <= yfsm_4;
          end 
        end
      endcase
    end
  end

  reg [32-1:0] zfsm;
  localparam zfsm_init = 0;
  localparam zfsm_1 = 1;
  localparam zfsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      zfsm <= zfsm_init;
    end else begin
      case(zfsm)
        zfsm_init: begin
          zready <= 0;
          if(reset_done) begin
            zfsm <= zfsm_1;
          end 
        end
        zfsm_1: begin
          zfsm <= zfsm_2;
        end
        zfsm_2: begin
          if(zvalid) begin
            zready <= 1;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(reset_done) begin
      $display("xdata=%d", xdata);
      $display("ydata=%d", ydata);
      $display("zdata=%d", zdata);
    end 
  end


endmodule



module main
(
  input CLK,
  input RST,
  input [4-1:0] xdata,
  input [4-1:0] ydata,
  output [4-1:0] zdata
);

  reg [1-1:0] _tmp_data_0;
  reg _tmp_valid_0;
  wire _tmp_ready_0;
  assign _tmp_ready_0 = (_tmp_ready_8 || !_tmp_valid_8) && (_tmp_valid_0 && _tmp_valid_1);
  reg [1-1:0] _tmp_data_1;
  reg _tmp_valid_1;
  wire _tmp_ready_1;
  assign _tmp_ready_1 = (_tmp_ready_8 || !_tmp_valid_8) && (_tmp_valid_0 && _tmp_valid_1);
  reg [1-1:0] _tmp_data_2;
  reg _tmp_valid_2;
  wire _tmp_ready_2;
  assign _tmp_ready_2 = (_tmp_ready_9 || !_tmp_valid_9) && _tmp_valid_2;
  reg [1-1:0] _tmp_data_3;
  reg _tmp_valid_3;
  wire _tmp_ready_3;
  assign _tmp_ready_3 = (_tmp_ready_10 || !_tmp_valid_10) && _tmp_valid_3;
  reg [1-1:0] _tmp_data_4;
  reg _tmp_valid_4;
  wire _tmp_ready_4;
  assign _tmp_ready_4 = (_tmp_ready_11 || !_tmp_valid_11) && _tmp_valid_4;
  reg [1-1:0] _tmp_data_5;
  reg _tmp_valid_5;
  wire _tmp_ready_5;
  assign _tmp_ready_5 = (_tmp_ready_12 || !_tmp_valid_12) && _tmp_valid_5;
  reg [1-1:0] _tmp_data_6;
  reg _tmp_valid_6;
  wire _tmp_ready_6;
  assign _tmp_ready_6 = (_tmp_ready_13 || !_tmp_valid_13) && _tmp_valid_6;
  reg [1-1:0] _tmp_data_7;
  reg _tmp_valid_7;
  wire _tmp_ready_7;
  assign _tmp_ready_7 = (_tmp_ready_14 || !_tmp_valid_14) && _tmp_valid_7;
  reg [1-1:0] _tmp_data_8;
  reg _tmp_valid_8;
  wire _tmp_ready_8;
  assign _tmp_ready_8 = (_tmp_ready_15 || !_tmp_valid_15) && (_tmp_valid_8 && _tmp_valid_9) && ((_tmp_ready_21 || !_tmp_valid_21) && _tmp_valid_8);
  reg [1-1:0] _tmp_data_9;
  reg _tmp_valid_9;
  wire _tmp_ready_9;
  assign _tmp_ready_9 = (_tmp_ready_15 || !_tmp_valid_15) && (_tmp_valid_8 && _tmp_valid_9);
  reg [1-1:0] _tmp_data_10;
  reg _tmp_valid_10;
  wire _tmp_ready_10;
  assign _tmp_ready_10 = (_tmp_ready_16 || !_tmp_valid_16) && _tmp_valid_10;
  reg [1-1:0] _tmp_data_11;
  reg _tmp_valid_11;
  wire _tmp_ready_11;
  assign _tmp_ready_11 = (_tmp_ready_17 || !_tmp_valid_17) && _tmp_valid_11;
  reg [1-1:0] _tmp_data_12;
  reg _tmp_valid_12;
  wire _tmp_ready_12;
  assign _tmp_ready_12 = (_tmp_ready_18 || !_tmp_valid_18) && _tmp_valid_12;
  reg [1-1:0] _tmp_data_13;
  reg _tmp_valid_13;
  wire _tmp_ready_13;
  assign _tmp_ready_13 = (_tmp_ready_19 || !_tmp_valid_19) && _tmp_valid_13;
  reg [1-1:0] _tmp_data_14;
  reg _tmp_valid_14;
  wire _tmp_ready_14;
  assign _tmp_ready_14 = (_tmp_ready_20 || !_tmp_valid_20) && _tmp_valid_14;
  reg [1-1:0] _tmp_data_15;
  reg _tmp_valid_15;
  wire _tmp_ready_15;
  assign _tmp_ready_15 = (_tmp_ready_22 || !_tmp_valid_22) && (_tmp_valid_15 && _tmp_valid_16);
  reg [1-1:0] _tmp_data_16;
  reg _tmp_valid_16;
  wire _tmp_ready_16;
  assign _tmp_ready_16 = (_tmp_ready_22 || !_tmp_valid_22) && (_tmp_valid_15 && _tmp_valid_16);
  reg [1-1:0] _tmp_data_17;
  reg _tmp_valid_17;
  wire _tmp_ready_17;
  assign _tmp_ready_17 = (_tmp_ready_23 || !_tmp_valid_23) && _tmp_valid_17;
  reg [1-1:0] _tmp_data_18;
  reg _tmp_valid_18;
  wire _tmp_ready_18;
  assign _tmp_ready_18 = (_tmp_ready_24 || !_tmp_valid_24) && _tmp_valid_18;
  reg [1-1:0] _tmp_data_19;
  reg _tmp_valid_19;
  wire _tmp_ready_19;
  assign _tmp_ready_19 = (_tmp_ready_25 || !_tmp_valid_25) && _tmp_valid_19;
  reg [1-1:0] _tmp_data_20;
  reg _tmp_valid_20;
  wire _tmp_ready_20;
  assign _tmp_ready_20 = (_tmp_ready_26 || !_tmp_valid_26) && _tmp_valid_20;
  reg [1-1:0] _tmp_data_21;
  reg _tmp_valid_21;
  wire _tmp_ready_21;
  assign _tmp_ready_21 = (_tmp_ready_27 || !_tmp_valid_27) && _tmp_valid_21;
  reg [1-1:0] _tmp_data_22;
  reg _tmp_valid_22;
  wire _tmp_ready_22;
  assign _tmp_ready_22 = (_tmp_ready_28 || !_tmp_valid_28) && (_tmp_valid_22 && _tmp_valid_23) && ((_tmp_ready_32 || !_tmp_valid_32) && _tmp_valid_22);
  reg [1-1:0] _tmp_data_23;
  reg _tmp_valid_23;
  wire _tmp_ready_23;
  assign _tmp_ready_23 = (_tmp_ready_28 || !_tmp_valid_28) && (_tmp_valid_22 && _tmp_valid_23);
  reg [1-1:0] _tmp_data_24;
  reg _tmp_valid_24;
  wire _tmp_ready_24;
  assign _tmp_ready_24 = (_tmp_ready_29 || !_tmp_valid_29) && _tmp_valid_24;
  reg [1-1:0] _tmp_data_25;
  reg _tmp_valid_25;
  wire _tmp_ready_25;
  assign _tmp_ready_25 = (_tmp_ready_30 || !_tmp_valid_30) && _tmp_valid_25;
  reg [1-1:0] _tmp_data_26;
  reg _tmp_valid_26;
  wire _tmp_ready_26;
  assign _tmp_ready_26 = (_tmp_ready_31 || !_tmp_valid_31) && _tmp_valid_26;
  reg [1-1:0] _tmp_data_27;
  reg _tmp_valid_27;
  wire _tmp_ready_27;
  assign _tmp_ready_27 = (_tmp_ready_33 || !_tmp_valid_33) && _tmp_valid_27;
  reg [1-1:0] _tmp_data_28;
  reg _tmp_valid_28;
  wire _tmp_ready_28;
  assign _tmp_ready_28 = (_tmp_ready_34 || !_tmp_valid_34) && (_tmp_valid_28 && _tmp_valid_29);
  reg [1-1:0] _tmp_data_29;
  reg _tmp_valid_29;
  wire _tmp_ready_29;
  assign _tmp_ready_29 = (_tmp_ready_34 || !_tmp_valid_34) && (_tmp_valid_28 && _tmp_valid_29);
  reg [1-1:0] _tmp_data_30;
  reg _tmp_valid_30;
  wire _tmp_ready_30;
  assign _tmp_ready_30 = (_tmp_ready_35 || !_tmp_valid_35) && _tmp_valid_30;
  reg [1-1:0] _tmp_data_31;
  reg _tmp_valid_31;
  wire _tmp_ready_31;
  assign _tmp_ready_31 = (_tmp_ready_36 || !_tmp_valid_36) && _tmp_valid_31;
  reg [1-1:0] _tmp_data_32;
  reg _tmp_valid_32;
  wire _tmp_ready_32;
  assign _tmp_ready_32 = (_tmp_ready_37 || !_tmp_valid_37) && _tmp_valid_32;
  reg [1-1:0] _tmp_data_33;
  reg _tmp_valid_33;
  wire _tmp_ready_33;
  assign _tmp_ready_33 = (_tmp_ready_38 || !_tmp_valid_38) && _tmp_valid_33;
  reg [1-1:0] _tmp_data_34;
  reg _tmp_valid_34;
  wire _tmp_ready_34;
  assign _tmp_ready_34 = (_tmp_ready_39 || !_tmp_valid_39) && (_tmp_valid_34 && _tmp_valid_35) && ((_tmp_ready_41 || !_tmp_valid_41) && _tmp_valid_34);
  reg [1-1:0] _tmp_data_35;
  reg _tmp_valid_35;
  wire _tmp_ready_35;
  assign _tmp_ready_35 = (_tmp_ready_39 || !_tmp_valid_39) && (_tmp_valid_34 && _tmp_valid_35);
  reg [1-1:0] _tmp_data_36;
  reg _tmp_valid_36;
  wire _tmp_ready_36;
  assign _tmp_ready_36 = (_tmp_ready_40 || !_tmp_valid_40) && _tmp_valid_36;
  reg [1-1:0] _tmp_data_37;
  reg _tmp_valid_37;
  wire _tmp_ready_37;
  assign _tmp_ready_37 = (_tmp_ready_42 || !_tmp_valid_42) && _tmp_valid_37;
  reg [1-1:0] _tmp_data_38;
  reg _tmp_valid_38;
  wire _tmp_ready_38;
  assign _tmp_ready_38 = (_tmp_ready_43 || !_tmp_valid_43) && _tmp_valid_38;
  reg [1-1:0] _tmp_data_39;
  reg _tmp_valid_39;
  wire _tmp_ready_39;
  assign _tmp_ready_39 = (_tmp_ready_44 || !_tmp_valid_44) && (_tmp_valid_39 && _tmp_valid_40);
  reg [1-1:0] _tmp_data_40;
  reg _tmp_valid_40;
  wire _tmp_ready_40;
  assign _tmp_ready_40 = (_tmp_ready_44 || !_tmp_valid_44) && (_tmp_valid_39 && _tmp_valid_40);
  reg [1-1:0] _tmp_data_41;
  reg _tmp_valid_41;
  wire _tmp_ready_41;
  assign _tmp_ready_41 = (_tmp_ready_45 || !_tmp_valid_45) && _tmp_valid_41;
  reg [1-1:0] _tmp_data_42;
  reg _tmp_valid_42;
  wire _tmp_ready_42;
  assign _tmp_ready_42 = (_tmp_ready_46 || !_tmp_valid_46) && _tmp_valid_42;
  reg [1-1:0] _tmp_data_43;
  reg _tmp_valid_43;
  wire _tmp_ready_43;
  assign _tmp_ready_43 = (_tmp_ready_47 || !_tmp_valid_47) && _tmp_valid_43;
  reg [1-1:0] _tmp_data_44;
  reg _tmp_valid_44;
  wire _tmp_ready_44;
  assign _tmp_ready_44 = (_tmp_ready_48 || !_tmp_valid_48) && (_tmp_valid_44 && _tmp_valid_45 && _tmp_valid_46 && _tmp_valid_47);
  reg [1-1:0] _tmp_data_45;
  reg _tmp_valid_45;
  wire _tmp_ready_45;
  assign _tmp_ready_45 = (_tmp_ready_48 || !_tmp_valid_48) && (_tmp_valid_44 && _tmp_valid_45 && _tmp_valid_46 && _tmp_valid_47);
  reg [1-1:0] _tmp_data_46;
  reg _tmp_valid_46;
  wire _tmp_ready_46;
  assign _tmp_ready_46 = (_tmp_ready_48 || !_tmp_valid_48) && (_tmp_valid_44 && _tmp_valid_45 && _tmp_valid_46 && _tmp_valid_47);
  reg [1-1:0] _tmp_data_47;
  reg _tmp_valid_47;
  wire _tmp_ready_47;
  assign _tmp_ready_47 = (_tmp_ready_48 || !_tmp_valid_48) && (_tmp_valid_44 && _tmp_valid_45 && _tmp_valid_46 && _tmp_valid_47);
  reg [4-1:0] _tmp_data_48;
  reg _tmp_valid_48;
  wire _tmp_ready_48;
  assign zdata = _tmp_data_48;
  assign _tmp_ready_48 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_0 <= 0;
      _tmp_valid_0 <= 0;
      _tmp_data_1 <= 0;
      _tmp_valid_1 <= 0;
      _tmp_data_2 <= 0;
      _tmp_valid_2 <= 0;
      _tmp_data_3 <= 0;
      _tmp_valid_3 <= 0;
      _tmp_data_4 <= 0;
      _tmp_valid_4 <= 0;
      _tmp_data_5 <= 0;
      _tmp_valid_5 <= 0;
      _tmp_data_6 <= 0;
      _tmp_valid_6 <= 0;
      _tmp_data_7 <= 0;
      _tmp_valid_7 <= 0;
      _tmp_data_8 <= 0;
      _tmp_valid_8 <= 0;
      _tmp_data_9 <= 0;
      _tmp_valid_9 <= 0;
      _tmp_data_10 <= 0;
      _tmp_valid_10 <= 0;
      _tmp_data_11 <= 0;
      _tmp_valid_11 <= 0;
      _tmp_data_12 <= 0;
      _tmp_valid_12 <= 0;
      _tmp_data_13 <= 0;
      _tmp_valid_13 <= 0;
      _tmp_data_14 <= 0;
      _tmp_valid_14 <= 0;
      _tmp_data_15 <= 0;
      _tmp_valid_15 <= 0;
      _tmp_data_16 <= 0;
      _tmp_valid_16 <= 0;
      _tmp_data_17 <= 0;
      _tmp_valid_17 <= 0;
      _tmp_data_18 <= 0;
      _tmp_valid_18 <= 0;
      _tmp_data_19 <= 0;
      _tmp_valid_19 <= 0;
      _tmp_data_20 <= 0;
      _tmp_valid_20 <= 0;
      _tmp_data_21 <= 0;
      _tmp_valid_21 <= 0;
      _tmp_data_22 <= 0;
      _tmp_valid_22 <= 0;
      _tmp_data_23 <= 0;
      _tmp_valid_23 <= 0;
      _tmp_data_24 <= 0;
      _tmp_valid_24 <= 0;
      _tmp_data_25 <= 0;
      _tmp_valid_25 <= 0;
      _tmp_data_26 <= 0;
      _tmp_valid_26 <= 0;
      _tmp_data_27 <= 0;
      _tmp_valid_27 <= 0;
      _tmp_data_28 <= 0;
      _tmp_valid_28 <= 0;
      _tmp_data_29 <= 0;
      _tmp_valid_29 <= 0;
      _tmp_data_30 <= 0;
      _tmp_valid_30 <= 0;
      _tmp_data_31 <= 0;
      _tmp_valid_31 <= 0;
      _tmp_data_32 <= 0;
      _tmp_valid_32 <= 0;
      _tmp_data_33 <= 0;
      _tmp_valid_33 <= 0;
      _tmp_data_34 <= 0;
      _tmp_valid_34 <= 0;
      _tmp_data_35 <= 0;
      _tmp_valid_35 <= 0;
      _tmp_data_36 <= 0;
      _tmp_valid_36 <= 0;
      _tmp_data_37 <= 0;
      _tmp_valid_37 <= 0;
      _tmp_data_38 <= 0;
      _tmp_valid_38 <= 0;
      _tmp_data_39 <= 0;
      _tmp_valid_39 <= 0;
      _tmp_data_40 <= 0;
      _tmp_valid_40 <= 0;
      _tmp_data_41 <= 0;
      _tmp_valid_41 <= 0;
      _tmp_data_42 <= 0;
      _tmp_valid_42 <= 0;
      _tmp_data_43 <= 0;
      _tmp_valid_43 <= 0;
      _tmp_data_44 <= 0;
      _tmp_valid_44 <= 0;
      _tmp_data_45 <= 0;
      _tmp_valid_45 <= 0;
      _tmp_data_46 <= 0;
      _tmp_valid_46 <= 0;
      _tmp_data_47 <= 0;
      _tmp_valid_47 <= 0;
      _tmp_data_48 <= 0;
      _tmp_valid_48 <= 0;
    end else begin
      if((_tmp_ready_0 || !_tmp_valid_0) && 1 && 1) begin
        _tmp_data_0 <= xdata[1'd0];
      end 
      if(_tmp_valid_0 && _tmp_ready_0) begin
        _tmp_valid_0 <= 0;
      end 
      if((_tmp_ready_0 || !_tmp_valid_0) && 1) begin
        _tmp_valid_0 <= 1;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1 && 1) begin
        _tmp_data_1 <= ydata[1'd0];
      end 
      if(_tmp_valid_1 && _tmp_ready_1) begin
        _tmp_valid_1 <= 0;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1) begin
        _tmp_valid_1 <= 1;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && 1 && 1) begin
        _tmp_data_2 <= xdata[2'd1];
      end 
      if(_tmp_valid_2 && _tmp_ready_2) begin
        _tmp_valid_2 <= 0;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && 1) begin
        _tmp_valid_2 <= 1;
      end 
      if((_tmp_ready_3 || !_tmp_valid_3) && 1 && 1) begin
        _tmp_data_3 <= ydata[2'd1];
      end 
      if(_tmp_valid_3 && _tmp_ready_3) begin
        _tmp_valid_3 <= 0;
      end 
      if((_tmp_ready_3 || !_tmp_valid_3) && 1) begin
        _tmp_valid_3 <= 1;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && 1 && 1) begin
        _tmp_data_4 <= xdata[3'd2];
      end 
      if(_tmp_valid_4 && _tmp_ready_4) begin
        _tmp_valid_4 <= 0;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && 1) begin
        _tmp_valid_4 <= 1;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && 1 && 1) begin
        _tmp_data_5 <= ydata[3'd2];
      end 
      if(_tmp_valid_5 && _tmp_ready_5) begin
        _tmp_valid_5 <= 0;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && 1) begin
        _tmp_valid_5 <= 1;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && 1 && 1) begin
        _tmp_data_6 <= xdata[3'd3];
      end 
      if(_tmp_valid_6 && _tmp_ready_6) begin
        _tmp_valid_6 <= 0;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && 1) begin
        _tmp_valid_6 <= 1;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && 1 && 1) begin
        _tmp_data_7 <= ydata[3'd3];
      end 
      if(_tmp_valid_7 && _tmp_ready_7) begin
        _tmp_valid_7 <= 0;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && 1) begin
        _tmp_valid_7 <= 1;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && (_tmp_ready_0 && _tmp_ready_1) && (_tmp_valid_0 && _tmp_valid_1)) begin
        _tmp_data_8 <= _tmp_data_0 ^ _tmp_data_1;
      end 
      if(_tmp_valid_8 && _tmp_ready_8) begin
        _tmp_valid_8 <= 0;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && (_tmp_ready_0 && _tmp_ready_1)) begin
        _tmp_valid_8 <= _tmp_valid_0 && _tmp_valid_1;
      end 
      if((_tmp_ready_9 || !_tmp_valid_9) && _tmp_ready_2 && _tmp_valid_2) begin
        _tmp_data_9 <= _tmp_data_2;
      end 
      if(_tmp_valid_9 && _tmp_ready_9) begin
        _tmp_valid_9 <= 0;
      end 
      if((_tmp_ready_9 || !_tmp_valid_9) && _tmp_ready_2) begin
        _tmp_valid_9 <= _tmp_valid_2;
      end 
      if((_tmp_ready_10 || !_tmp_valid_10) && _tmp_ready_3 && _tmp_valid_3) begin
        _tmp_data_10 <= _tmp_data_3;
      end 
      if(_tmp_valid_10 && _tmp_ready_10) begin
        _tmp_valid_10 <= 0;
      end 
      if((_tmp_ready_10 || !_tmp_valid_10) && _tmp_ready_3) begin
        _tmp_valid_10 <= _tmp_valid_3;
      end 
      if((_tmp_ready_11 || !_tmp_valid_11) && _tmp_ready_4 && _tmp_valid_4) begin
        _tmp_data_11 <= _tmp_data_4;
      end 
      if(_tmp_valid_11 && _tmp_ready_11) begin
        _tmp_valid_11 <= 0;
      end 
      if((_tmp_ready_11 || !_tmp_valid_11) && _tmp_ready_4) begin
        _tmp_valid_11 <= _tmp_valid_4;
      end 
      if((_tmp_ready_12 || !_tmp_valid_12) && _tmp_ready_5 && _tmp_valid_5) begin
        _tmp_data_12 <= _tmp_data_5;
      end 
      if(_tmp_valid_12 && _tmp_ready_12) begin
        _tmp_valid_12 <= 0;
      end 
      if((_tmp_ready_12 || !_tmp_valid_12) && _tmp_ready_5) begin
        _tmp_valid_12 <= _tmp_valid_5;
      end 
      if((_tmp_ready_13 || !_tmp_valid_13) && _tmp_ready_6 && _tmp_valid_6) begin
        _tmp_data_13 <= _tmp_data_6;
      end 
      if(_tmp_valid_13 && _tmp_ready_13) begin
        _tmp_valid_13 <= 0;
      end 
      if((_tmp_ready_13 || !_tmp_valid_13) && _tmp_ready_6) begin
        _tmp_valid_13 <= _tmp_valid_6;
      end 
      if((_tmp_ready_14 || !_tmp_valid_14) && _tmp_ready_7 && _tmp_valid_7) begin
        _tmp_data_14 <= _tmp_data_7;
      end 
      if(_tmp_valid_14 && _tmp_ready_14) begin
        _tmp_valid_14 <= 0;
      end 
      if((_tmp_ready_14 || !_tmp_valid_14) && _tmp_ready_7) begin
        _tmp_valid_14 <= _tmp_valid_7;
      end 
      if((_tmp_ready_15 || !_tmp_valid_15) && (_tmp_ready_8 && _tmp_ready_9) && (_tmp_valid_8 && _tmp_valid_9)) begin
        _tmp_data_15 <= _tmp_data_8 ^ _tmp_data_9;
      end 
      if(_tmp_valid_15 && _tmp_ready_15) begin
        _tmp_valid_15 <= 0;
      end 
      if((_tmp_ready_15 || !_tmp_valid_15) && (_tmp_ready_8 && _tmp_ready_9)) begin
        _tmp_valid_15 <= _tmp_valid_8 && _tmp_valid_9;
      end 
      if((_tmp_ready_16 || !_tmp_valid_16) && _tmp_ready_10 && _tmp_valid_10) begin
        _tmp_data_16 <= _tmp_data_10;
      end 
      if(_tmp_valid_16 && _tmp_ready_16) begin
        _tmp_valid_16 <= 0;
      end 
      if((_tmp_ready_16 || !_tmp_valid_16) && _tmp_ready_10) begin
        _tmp_valid_16 <= _tmp_valid_10;
      end 
      if((_tmp_ready_17 || !_tmp_valid_17) && _tmp_ready_11 && _tmp_valid_11) begin
        _tmp_data_17 <= _tmp_data_11;
      end 
      if(_tmp_valid_17 && _tmp_ready_17) begin
        _tmp_valid_17 <= 0;
      end 
      if((_tmp_ready_17 || !_tmp_valid_17) && _tmp_ready_11) begin
        _tmp_valid_17 <= _tmp_valid_11;
      end 
      if((_tmp_ready_18 || !_tmp_valid_18) && _tmp_ready_12 && _tmp_valid_12) begin
        _tmp_data_18 <= _tmp_data_12;
      end 
      if(_tmp_valid_18 && _tmp_ready_18) begin
        _tmp_valid_18 <= 0;
      end 
      if((_tmp_ready_18 || !_tmp_valid_18) && _tmp_ready_12) begin
        _tmp_valid_18 <= _tmp_valid_12;
      end 
      if((_tmp_ready_19 || !_tmp_valid_19) && _tmp_ready_13 && _tmp_valid_13) begin
        _tmp_data_19 <= _tmp_data_13;
      end 
      if(_tmp_valid_19 && _tmp_ready_19) begin
        _tmp_valid_19 <= 0;
      end 
      if((_tmp_ready_19 || !_tmp_valid_19) && _tmp_ready_13) begin
        _tmp_valid_19 <= _tmp_valid_13;
      end 
      if((_tmp_ready_20 || !_tmp_valid_20) && _tmp_ready_14 && _tmp_valid_14) begin
        _tmp_data_20 <= _tmp_data_14;
      end 
      if(_tmp_valid_20 && _tmp_ready_20) begin
        _tmp_valid_20 <= 0;
      end 
      if((_tmp_ready_20 || !_tmp_valid_20) && _tmp_ready_14) begin
        _tmp_valid_20 <= _tmp_valid_14;
      end 
      if((_tmp_ready_21 || !_tmp_valid_21) && _tmp_ready_8 && _tmp_valid_8) begin
        _tmp_data_21 <= _tmp_data_8;
      end 
      if(_tmp_valid_21 && _tmp_ready_21) begin
        _tmp_valid_21 <= 0;
      end 
      if((_tmp_ready_21 || !_tmp_valid_21) && _tmp_ready_8) begin
        _tmp_valid_21 <= _tmp_valid_8;
      end 
      if((_tmp_ready_22 || !_tmp_valid_22) && (_tmp_ready_15 && _tmp_ready_16) && (_tmp_valid_15 && _tmp_valid_16)) begin
        _tmp_data_22 <= _tmp_data_15 ^ _tmp_data_16;
      end 
      if(_tmp_valid_22 && _tmp_ready_22) begin
        _tmp_valid_22 <= 0;
      end 
      if((_tmp_ready_22 || !_tmp_valid_22) && (_tmp_ready_15 && _tmp_ready_16)) begin
        _tmp_valid_22 <= _tmp_valid_15 && _tmp_valid_16;
      end 
      if((_tmp_ready_23 || !_tmp_valid_23) && _tmp_ready_17 && _tmp_valid_17) begin
        _tmp_data_23 <= _tmp_data_17;
      end 
      if(_tmp_valid_23 && _tmp_ready_23) begin
        _tmp_valid_23 <= 0;
      end 
      if((_tmp_ready_23 || !_tmp_valid_23) && _tmp_ready_17) begin
        _tmp_valid_23 <= _tmp_valid_17;
      end 
      if((_tmp_ready_24 || !_tmp_valid_24) && _tmp_ready_18 && _tmp_valid_18) begin
        _tmp_data_24 <= _tmp_data_18;
      end 
      if(_tmp_valid_24 && _tmp_ready_24) begin
        _tmp_valid_24 <= 0;
      end 
      if((_tmp_ready_24 || !_tmp_valid_24) && _tmp_ready_18) begin
        _tmp_valid_24 <= _tmp_valid_18;
      end 
      if((_tmp_ready_25 || !_tmp_valid_25) && _tmp_ready_19 && _tmp_valid_19) begin
        _tmp_data_25 <= _tmp_data_19;
      end 
      if(_tmp_valid_25 && _tmp_ready_25) begin
        _tmp_valid_25 <= 0;
      end 
      if((_tmp_ready_25 || !_tmp_valid_25) && _tmp_ready_19) begin
        _tmp_valid_25 <= _tmp_valid_19;
      end 
      if((_tmp_ready_26 || !_tmp_valid_26) && _tmp_ready_20 && _tmp_valid_20) begin
        _tmp_data_26 <= _tmp_data_20;
      end 
      if(_tmp_valid_26 && _tmp_ready_26) begin
        _tmp_valid_26 <= 0;
      end 
      if((_tmp_ready_26 || !_tmp_valid_26) && _tmp_ready_20) begin
        _tmp_valid_26 <= _tmp_valid_20;
      end 
      if((_tmp_ready_27 || !_tmp_valid_27) && _tmp_ready_21 && _tmp_valid_21) begin
        _tmp_data_27 <= _tmp_data_21;
      end 
      if(_tmp_valid_27 && _tmp_ready_27) begin
        _tmp_valid_27 <= 0;
      end 
      if((_tmp_ready_27 || !_tmp_valid_27) && _tmp_ready_21) begin
        _tmp_valid_27 <= _tmp_valid_21;
      end 
      if((_tmp_ready_28 || !_tmp_valid_28) && (_tmp_ready_22 && _tmp_ready_23) && (_tmp_valid_22 && _tmp_valid_23)) begin
        _tmp_data_28 <= _tmp_data_22 ^ _tmp_data_23;
      end 
      if(_tmp_valid_28 && _tmp_ready_28) begin
        _tmp_valid_28 <= 0;
      end 
      if((_tmp_ready_28 || !_tmp_valid_28) && (_tmp_ready_22 && _tmp_ready_23)) begin
        _tmp_valid_28 <= _tmp_valid_22 && _tmp_valid_23;
      end 
      if((_tmp_ready_29 || !_tmp_valid_29) && _tmp_ready_24 && _tmp_valid_24) begin
        _tmp_data_29 <= _tmp_data_24;
      end 
      if(_tmp_valid_29 && _tmp_ready_29) begin
        _tmp_valid_29 <= 0;
      end 
      if((_tmp_ready_29 || !_tmp_valid_29) && _tmp_ready_24) begin
        _tmp_valid_29 <= _tmp_valid_24;
      end 
      if((_tmp_ready_30 || !_tmp_valid_30) && _tmp_ready_25 && _tmp_valid_25) begin
        _tmp_data_30 <= _tmp_data_25;
      end 
      if(_tmp_valid_30 && _tmp_ready_30) begin
        _tmp_valid_30 <= 0;
      end 
      if((_tmp_ready_30 || !_tmp_valid_30) && _tmp_ready_25) begin
        _tmp_valid_30 <= _tmp_valid_25;
      end 
      if((_tmp_ready_31 || !_tmp_valid_31) && _tmp_ready_26 && _tmp_valid_26) begin
        _tmp_data_31 <= _tmp_data_26;
      end 
      if(_tmp_valid_31 && _tmp_ready_31) begin
        _tmp_valid_31 <= 0;
      end 
      if((_tmp_ready_31 || !_tmp_valid_31) && _tmp_ready_26) begin
        _tmp_valid_31 <= _tmp_valid_26;
      end 
      if((_tmp_ready_32 || !_tmp_valid_32) && _tmp_ready_22 && _tmp_valid_22) begin
        _tmp_data_32 <= _tmp_data_22;
      end 
      if(_tmp_valid_32 && _tmp_ready_32) begin
        _tmp_valid_32 <= 0;
      end 
      if((_tmp_ready_32 || !_tmp_valid_32) && _tmp_ready_22) begin
        _tmp_valid_32 <= _tmp_valid_22;
      end 
      if((_tmp_ready_33 || !_tmp_valid_33) && _tmp_ready_27 && _tmp_valid_27) begin
        _tmp_data_33 <= _tmp_data_27;
      end 
      if(_tmp_valid_33 && _tmp_ready_33) begin
        _tmp_valid_33 <= 0;
      end 
      if((_tmp_ready_33 || !_tmp_valid_33) && _tmp_ready_27) begin
        _tmp_valid_33 <= _tmp_valid_27;
      end 
      if((_tmp_ready_34 || !_tmp_valid_34) && (_tmp_ready_28 && _tmp_ready_29) && (_tmp_valid_28 && _tmp_valid_29)) begin
        _tmp_data_34 <= _tmp_data_28 ^ _tmp_data_29;
      end 
      if(_tmp_valid_34 && _tmp_ready_34) begin
        _tmp_valid_34 <= 0;
      end 
      if((_tmp_ready_34 || !_tmp_valid_34) && (_tmp_ready_28 && _tmp_ready_29)) begin
        _tmp_valid_34 <= _tmp_valid_28 && _tmp_valid_29;
      end 
      if((_tmp_ready_35 || !_tmp_valid_35) && _tmp_ready_30 && _tmp_valid_30) begin
        _tmp_data_35 <= _tmp_data_30;
      end 
      if(_tmp_valid_35 && _tmp_ready_35) begin
        _tmp_valid_35 <= 0;
      end 
      if((_tmp_ready_35 || !_tmp_valid_35) && _tmp_ready_30) begin
        _tmp_valid_35 <= _tmp_valid_30;
      end 
      if((_tmp_ready_36 || !_tmp_valid_36) && _tmp_ready_31 && _tmp_valid_31) begin
        _tmp_data_36 <= _tmp_data_31;
      end 
      if(_tmp_valid_36 && _tmp_ready_36) begin
        _tmp_valid_36 <= 0;
      end 
      if((_tmp_ready_36 || !_tmp_valid_36) && _tmp_ready_31) begin
        _tmp_valid_36 <= _tmp_valid_31;
      end 
      if((_tmp_ready_37 || !_tmp_valid_37) && _tmp_ready_32 && _tmp_valid_32) begin
        _tmp_data_37 <= _tmp_data_32;
      end 
      if(_tmp_valid_37 && _tmp_ready_37) begin
        _tmp_valid_37 <= 0;
      end 
      if((_tmp_ready_37 || !_tmp_valid_37) && _tmp_ready_32) begin
        _tmp_valid_37 <= _tmp_valid_32;
      end 
      if((_tmp_ready_38 || !_tmp_valid_38) && _tmp_ready_33 && _tmp_valid_33) begin
        _tmp_data_38 <= _tmp_data_33;
      end 
      if(_tmp_valid_38 && _tmp_ready_38) begin
        _tmp_valid_38 <= 0;
      end 
      if((_tmp_ready_38 || !_tmp_valid_38) && _tmp_ready_33) begin
        _tmp_valid_38 <= _tmp_valid_33;
      end 
      if((_tmp_ready_39 || !_tmp_valid_39) && (_tmp_ready_34 && _tmp_ready_35) && (_tmp_valid_34 && _tmp_valid_35)) begin
        _tmp_data_39 <= _tmp_data_34 ^ _tmp_data_35;
      end 
      if(_tmp_valid_39 && _tmp_ready_39) begin
        _tmp_valid_39 <= 0;
      end 
      if((_tmp_ready_39 || !_tmp_valid_39) && (_tmp_ready_34 && _tmp_ready_35)) begin
        _tmp_valid_39 <= _tmp_valid_34 && _tmp_valid_35;
      end 
      if((_tmp_ready_40 || !_tmp_valid_40) && _tmp_ready_36 && _tmp_valid_36) begin
        _tmp_data_40 <= _tmp_data_36;
      end 
      if(_tmp_valid_40 && _tmp_ready_40) begin
        _tmp_valid_40 <= 0;
      end 
      if((_tmp_ready_40 || !_tmp_valid_40) && _tmp_ready_36) begin
        _tmp_valid_40 <= _tmp_valid_36;
      end 
      if((_tmp_ready_41 || !_tmp_valid_41) && _tmp_ready_34 && _tmp_valid_34) begin
        _tmp_data_41 <= _tmp_data_34;
      end 
      if(_tmp_valid_41 && _tmp_ready_41) begin
        _tmp_valid_41 <= 0;
      end 
      if((_tmp_ready_41 || !_tmp_valid_41) && _tmp_ready_34) begin
        _tmp_valid_41 <= _tmp_valid_34;
      end 
      if((_tmp_ready_42 || !_tmp_valid_42) && _tmp_ready_37 && _tmp_valid_37) begin
        _tmp_data_42 <= _tmp_data_37;
      end 
      if(_tmp_valid_42 && _tmp_ready_42) begin
        _tmp_valid_42 <= 0;
      end 
      if((_tmp_ready_42 || !_tmp_valid_42) && _tmp_ready_37) begin
        _tmp_valid_42 <= _tmp_valid_37;
      end 
      if((_tmp_ready_43 || !_tmp_valid_43) && _tmp_ready_38 && _tmp_valid_38) begin
        _tmp_data_43 <= _tmp_data_38;
      end 
      if(_tmp_valid_43 && _tmp_ready_43) begin
        _tmp_valid_43 <= 0;
      end 
      if((_tmp_ready_43 || !_tmp_valid_43) && _tmp_ready_38) begin
        _tmp_valid_43 <= _tmp_valid_38;
      end 
      if((_tmp_ready_44 || !_tmp_valid_44) && (_tmp_ready_39 && _tmp_ready_40) && (_tmp_valid_39 && _tmp_valid_40)) begin
        _tmp_data_44 <= _tmp_data_39 ^ _tmp_data_40;
      end 
      if(_tmp_valid_44 && _tmp_ready_44) begin
        _tmp_valid_44 <= 0;
      end 
      if((_tmp_ready_44 || !_tmp_valid_44) && (_tmp_ready_39 && _tmp_ready_40)) begin
        _tmp_valid_44 <= _tmp_valid_39 && _tmp_valid_40;
      end 
      if((_tmp_ready_45 || !_tmp_valid_45) && _tmp_ready_41 && _tmp_valid_41) begin
        _tmp_data_45 <= _tmp_data_41;
      end 
      if(_tmp_valid_45 && _tmp_ready_45) begin
        _tmp_valid_45 <= 0;
      end 
      if((_tmp_ready_45 || !_tmp_valid_45) && _tmp_ready_41) begin
        _tmp_valid_45 <= _tmp_valid_41;
      end 
      if((_tmp_ready_46 || !_tmp_valid_46) && _tmp_ready_42 && _tmp_valid_42) begin
        _tmp_data_46 <= _tmp_data_42;
      end 
      if(_tmp_valid_46 && _tmp_ready_46) begin
        _tmp_valid_46 <= 0;
      end 
      if((_tmp_ready_46 || !_tmp_valid_46) && _tmp_ready_42) begin
        _tmp_valid_46 <= _tmp_valid_42;
      end 
      if((_tmp_ready_47 || !_tmp_valid_47) && _tmp_ready_43 && _tmp_valid_43) begin
        _tmp_data_47 <= _tmp_data_43;
      end 
      if(_tmp_valid_47 && _tmp_ready_47) begin
        _tmp_valid_47 <= 0;
      end 
      if((_tmp_ready_47 || !_tmp_valid_47) && _tmp_ready_43) begin
        _tmp_valid_47 <= _tmp_valid_43;
      end 
      if((_tmp_ready_48 || !_tmp_valid_48) && (_tmp_ready_44 && _tmp_ready_45 && _tmp_ready_46 && _tmp_ready_47) && (_tmp_valid_44 && _tmp_valid_45 && _tmp_valid_46 && _tmp_valid_47)) begin
        _tmp_data_48 <= { _tmp_data_44, _tmp_data_45, _tmp_data_46, _tmp_data_47 };
      end 
      if(_tmp_valid_48 && _tmp_ready_48) begin
        _tmp_valid_48 <= 0;
      end 
      if((_tmp_ready_48 || !_tmp_valid_48) && (_tmp_ready_44 && _tmp_ready_45 && _tmp_ready_46 && _tmp_ready_47)) begin
        _tmp_valid_48 <= _tmp_valid_44 && _tmp_valid_45 && _tmp_valid_46 && _tmp_valid_47;
      end 
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = dataflow__iter.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import copy

import veriloggen.core.vtypes as vtypes


class RenameVisitor(object):

    def __init__(self, prefix=None, postfix=None):
        self.prefix = prefix if prefix is not None else ''
        self.postfix = postfix if postfix is not None else ''

    def generic_visit(self, node):
        raise TypeError("Type %s is not supported." % str(type(node)))

    def visit(self, node):
        if node is None:
            return None
        if isinstance(node, vtypes._Variable):
            return self.visit__Variable(node)
        if isinstance(node, vtypes._BinaryOperator):
            return self.visit__BinaryOperator(node)
        if isinstance(node, vtypes._UnaryOperator):
            return self.visit__UnaryOperator(node)

        visitor = getattr(
            self, 'visit_' + node.__class__.__name__, self.generic_visit)
        return visitor(node)

    def visit_Int(self, node):
        return node

    def visit_Float(self, node):
        return node

    def visit_Str(self, node):
        return node

    def visit_bool(self, node):
        return node

    def visit_int(self, node):
        return node

    def visit_str(self, node):
        return node

    def visit_float(self, node):
        return node

    def visit__Variable(self, node):
        ret = copy.deepcopy(node)
        ret.name = ''.join([self.prefix, ret.name, self.postfix])
        return ret

    def visit_Pointer(self, node):
        var = self.visit(node.var)
        pos = self.visit(node.pos)
        return vtypes.Pointer(var, pos)

    def visit_Slice(self, node):
        var = self.visit(node.var)
        msb = self.visit(node.msb)
        lsb = self.visit(node.lsb)
        return vtypes.Slice(var, msb, lsb)

    def visit_Cat(self, node):
        vars = [self.visit(var) for var in node.vars]
        return vtypes.Cat(*vars)

    def visit_Repeat(self, node):
        var = self.visit(node.var)
        times = self.visit(node.times)
        return vtypes.Repeat(var, times)

    def visit_Cond(self, node):
        condition = self.visit(node.condition)
        true_value = self.visit(node.true_value)
        false_value = self.visit(node.false_value)
        return vtype.Cond(condition, true_value, false_value)

    def visit__BinaryOperator(self, node):
        op = type(node)
        left = self.visit(node.left)
        right = self.visit(node.right)
        return op(left, right)

    def visit__UnaryOperator(self, node):
        op = type(node)
        right = self.visit(node.right)
        return op(right)
